#include <string>
#include <queue>
#include <list>
#include <vector>
#include <iostream>
#include <pthread.h>
#include <unistd.h>
#include <fstream>

#define STRING	std::string
#define LIST	std::list<int>
#define CIN	std::cin
#define COUT	std::cout
#define ENDL	std::endl
#define QUEUE	std::list<PCB>
#define VECTOR	std::vector<PCB>

pthread_mutex_t cpu_mutex = PTHREAD_MUTEX_INITIALIZER;

std::ofstream logfile;

class PCB
{
public:
	PCB(LIST times, STRING name, int pid)
	{
		this->name = name;
		this->times = times;
		this->pid = pid;
	}

	int getPid(){ return pid; }
	STRING getName(){ return name; }
	
	bool isEnd()
	{
		if(times.size() == 0)
			return true;
		
		return false;
	}
	
	bool isIO()
	{ 
		if((times.size() % 2) == 0)
			return true;
		
		return false;
	}
	
	void oneBurst()
	{ 
		int i = times.front();
		times.pop_front();
		i--;

		if( i != 0)		
			times.push_front(i);		
	}

private:
	STRING name;
	LIST times;
	int pid;
};

class CPU
{
public:
	CPU(int number)
	{
		this->number = number;
		freePid = 1;
		ended = false;
		idle = 0;
		wait = 0;
		response = 0;
		cycles = 0;
	}

	int addProcess(LIST bursts, STRING name)
	{
		Jobs.push_back(PCB(bursts, name, freePid++));

		return (freePid - 1);
	}

	bool isEnded(){ return ended; }
	void setEnded(){ ended = true; }
	float getWait(){ return float(wait)/float(freePid - 1); }
	float getResponse(){ return float(response)/float(freePid - 1); }
	
	int getIdle()
	{
		if( cycles == 0)
			return 100;

		float retval = float(idle)/float(cycles);
		
		return int(retval * 100);
	}
	
	STRING findProcess(STRING name)
	{
		STRING retval = "The Process Not Found !!\n";
		
		QUEUE::iterator it = Jobs.begin();

		if(it != Jobs.end() && it->getName() == name)
		{
			retval  = "Name\tPID\tState\n";
			retval += it->getName();
			retval += "\t";
			retval += int2str(it->getPid());
			retval += "\trunning\n";
			return retval;
		}
		
		for(it++; it != Jobs.end(); it++)

		{
			if(it->getName() == name)
			{
				retval  = "Name\tPID\tState\n";
				retval += it->getName();
				retval += "\t";
				retval += int2str(it->getPid());
				retval += "\tready\n";
				return retval;
			}
		}

		for(int i = IOs.size() - 1; i >= 0; i--)
		{
			if(IOs[i].getName() == name)
			{
				retval  = "Name\tPID\tState\n";
				retval += IOs[i].getName();
				retval += "\t";
				retval += int2str(IOs[i].getPid());
				retval += "\twaiting\n";
				return retval;
			}
		}

		return retval;
	}
	
	bool kill(STRING name)
	{
		for(QUEUE::iterator it = Jobs.begin(); it != Jobs.end(); it++)
		{
			if(it->getName() == name)
			{
				Jobs.erase(it);
				return true;
			}
		}

		for(VECTOR::iterator it = IOs.begin(); it != IOs.end(); it++)
		{
			if(it->getName() == name)
			{
				IOs.erase(it);
				return true;
			}
		}

		return false;
	}

	bool contains(STRING name)
	{
		for(QUEUE::iterator it = Jobs.begin(); it != Jobs.end(); it++)
		{
			if(it->getName() == name)
				return true;
		}

		for(int i = IOs.size() - 1; i >= 0; i--)
		{
			if(IOs[i].getName() == name)
				return true;
		}

		return false;
	}
	
	STRING toString()
	{
		STRING retval = "Name\tPID\tState\n";
		
		QUEUE::iterator it = Jobs.begin();

		if(it != Jobs.end())
		{
			retval += it->getName();
			retval += "\t";
			retval += int2str(it->getPid());
			retval += "\trunning\n";
			it++;
		}
		
		for(; it != Jobs.end(); it++)
		{
			retval += it->getName();
			retval += "\t";
			retval += int2str(it->getPid());
			retval += "\tready\n";
		}

		for(int i = IOs.size() - 1; i >= 0; i--)
		{
			retval += IOs[i].getName();
			retval += "\t";
			retval += int2str(IOs[i].getPid());
			retval += "\twaiting\n";
		}

		return retval;
	}

	void oneCycle()
	{
		cycles++;
		bool flag = false;

		if(Jobs.empty())
		{
			idle++;
			logfile << "CPU is idle." << ENDL;
		}
		else
		{
			QUEUE::iterator it = Jobs.begin();
			it->oneBurst();
			logfile << "One of the bursts of the process with Pid: " << it->getPid() << " has been done." << ENDL;
			response++;
			it++;
		
			for(; it != Jobs.end(); it++)
			{
				response++;
				wait++;
			}

			flag = true;
		}

		for(int i = 0; i < IOs.size(); i++)
		{
			IOs[i].oneBurst();
			logfile << "One of the IO bursts of the process with Pid: " << IOs[i].getPid() << " has been done." << ENDL;

			if(!IOs[i].isIO())
			{
				Jobs.push_back(IOs[i]);

				logfile << "The process with Pid: " << IOs[i].getPid() << " has been moved to ready queue after doing IO." << ENDL;
				
				VECTOR::iterator vit = IOs.begin();
				for(int j = 0; j < i; j++)
					vit++;
				IOs.erase(vit);

				i--;
			}	
		}

		if(flag)
		{
			if(!Jobs.begin()->isEnd())
			{
				if(!Jobs.begin()->isIO())
					Jobs.push_back(*(Jobs.begin()));
				else
					IOs.push_back(*(Jobs.begin()));
			}
			else
				logfile << "The process with Pid: " << Jobs.begin()->getPid() << " has been ended." << ENDL;
				
			Jobs.erase(Jobs.begin());
		}
	}

	int getNumber(){ return number; }

private:
	STRING int2str(int num)
	{
		STRING temp;

		if(num <= 0)
			return "0";

		for(;num != 0; num /= 10)
			temp += '0' + num - ((num / 10) * 10);

		STRING retval;

		for(int i = temp.size() - 1; i >= 0; i--)
			retval += temp[i];

		return retval;
	}

	int idle;
	int wait;
	int response;
	int cycles;

	int freePid;
	int number;
	bool ended;
	QUEUE Jobs;
	VECTOR IOs;
};

void *runner(void *parameter)
{
	CPU * mycpu = (CPU *)parameter;

	pthread_mutex_lock(&cpu_mutex);
	bool ended = mycpu->isEnded();
	pthread_mutex_unlock(&cpu_mutex);

	while(!ended)
	{
		pthread_mutex_lock(&cpu_mutex);
		mycpu->oneCycle();
		pthread_mutex_unlock(&cpu_mutex);
	
		sleep(5);

		pthread_mutex_lock(&cpu_mutex);
		ended = mycpu->isEnded();
		pthread_mutex_unlock(&cpu_mutex);
	}

	return NULL;
}

int main()
{
	int args;
	char command[256];
	STRING allcommands[] = { "Create", "Ps", "Kill", "Quit"};
	char * argv[256];
	command[0] = 0;
	logfile.open("logfile");
	CPU mycpu(1);

	pthread_t thread_id;
	pthread_create(&thread_id, NULL, &runner, &mycpu);

	for(;;)
	{
		COUT << "/> ";
		CIN.getline( command, 255);
		
		while(command[0] == 0)
			CIN.getline( command, 255);
		
		args = 0;

		for( int ind = 0; command[ind] != 0;)
		{
			argv[args] = command + ind;			
				
			while( command[ind] != ' ' && command[ind] != 0)
				ind++;

			if( command[ind] != 0)
			{
				command[ind] = 0;
				ind++;
			}

			args++;	
			argv[args] = NULL;
		}
	
		if( argv[0] == NULL)
			continue;

		if( allcommands[0] == argv[0])
		{
			if( args == 2)
			{
				pthread_mutex_lock(&cpu_mutex);
				bool contains = mycpu.contains(STRING(argv[1]));
				pthread_mutex_unlock(&cpu_mutex);			

				if(contains)
					COUT << "This proccess already exists !!" << ENDL;
				else
				{
					int num;
					int temp;
					LIST bursts;
					COUT << "Please Enter Number of CPU and IO Bursts: ";
					CIN >> num;
				
					if( num % 2 == 0)
						COUT << "Number of bursts must be odd !!" << ENDL;
					else{
						for(int i = 0; i < num; i++)
						{
							COUT << "\t: ";
							CIN >> temp;
							bursts.push_back(temp);
						}					

						COUT << "The Process Added by PID ";
						pthread_mutex_lock(&cpu_mutex);
						COUT << mycpu.addProcess(bursts, STRING(argv[1])) << ENDL;
						pthread_mutex_unlock(&cpu_mutex);
					}
				}
			}
			else
				COUT << "The instruction gets 1 parameter !!" << ENDL;
		}
		else if( allcommands[1] == argv[0])
		{
			if( args == 1)
			{
				pthread_mutex_lock(&cpu_mutex);
				STRING result = mycpu.toString();
				pthread_mutex_unlock(&cpu_mutex);
				COUT << result;
			}
			else if( args == 2)
			{
				pthread_mutex_lock(&cpu_mutex);
				STRING result = mycpu.findProcess(argv[1]);
				pthread_mutex_unlock(&cpu_mutex);
				COUT << result;
			}
			else
				COUT << "The instruction gets 1 or 0 parameter !!" << ENDL;
		}
		else if( allcommands[2] == argv[0])
		{	
			if( args == 2)
			{
				pthread_mutex_lock(&cpu_mutex);
				bool kill = mycpu.kill(STRING(argv[1]));
				pthread_mutex_unlock(&cpu_mutex);

				if(kill)
					COUT << "Process " << argv[1] << " Terminated." << ENDL;
				else
					COUT << "The Process Not Found !!" << ENDL;
			}
			else				
				COUT << "The instruction gets 1 parameter !!" << ENDL;		
		}
		else if( allcommands[3] == argv[0])
		{
			if( args == 1)
			{
				pthread_mutex_lock(&cpu_mutex);
				mycpu.setEnded();
				pthread_mutex_unlock(&cpu_mutex);
				break;
			}
			else				
				COUT << "The instruction gets 0 parameter !!" << ENDL;
		}
		else
		{
			COUT << "Unknown command !!" << ENDL;
			continue;
		}
	}

	pthread_join(thread_id, NULL);

	COUT << "\tCPU Idle = " << mycpu.getIdle() << "%" << ENDL;
	COUT << "\tAverage Waiting Time = " << mycpu.getWait() << ENDL;
	COUT << "\tAverage Response Time = " << mycpu.getResponse() << ENDL;

	logfile.close();
	return 0;
}


